-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation: Healing")

L:RegisterTranslations("enUS", function() return {
	["Healing done"] = true,
	["H"] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Healing done"] = "치유량",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local playerName = nil

local data = nil
local broadcastdirty = 0
local dirty = nil
local moduleName = "Healing"

local parser = ParserLib:GetInstance("1.1")
local roster = AceLibrary("Roster-2.1")

-----------------------------------------------------------------------
-- Module declaration
-----------------------------------------------------------------------

local mod = Violation:NewModule(moduleName)
mod.revision = tonumber(("$Revision: 36880 $"):sub(12, -3)) or 1
mod.version = 4
mod.canMergeData = true

-----------------------------------------------------------------------
-- Module initialization
-----------------------------------------------------------------------

function mod:OnModuleRegister()
	playerName = UnitName("player")
	data = self:new()
end

function mod:OnModuleEnable()
	parser:RegisterInfoType( "Violation "..moduleName, "heal", function( infoType, info ) self:OnParserEvent( infoType, info ) end )
end

function mod:OnModuleDisable()
	parser:UnregisterAllInfoTypes( "Violation "..moduleName )
end

function mod:OnReset()
	if type(data) == "table" then
		data = self:del(data)
	end
	data = self:new()
	data[playerName] = 0
	broadcastdirty = 0
	dirty = nil
end

-----------------------------------------------------------------------
-- Module implementation
-----------------------------------------------------------------------

function mod:GetRevision() return self.revision end
function mod:GetDisplayName() return L["Healing done"] end
function mod:GetShortDisplayName() return L["H"] end
function mod:GetBroadcastData()
	local x = broadcastdirty
	broadcastdirty = 0
	return x
end

function mod:IsBroadcastDirty()
	return broadcastdirty > 0 and true or false
end

function mod:IsDirty()
	return dirty
end

function mod:ResetDirty()
	dirty = nil
end

function mod:OnRemoteData(sender, input)
	data[sender] = (data[sender] or 0) + input
	dirty = true
end

local function getOverheal( unit, amount )
	local overheal = 0
	if unit and UnitExists(unit) then
		local d = UnitHealthMax(unit)-UnitHealth(unit)
		if d < amount then
			overheal = amount - d
		end
	end
	return overheal
end

function mod:OnParserEvent( infoType, info )
	if infoType ~= "heal" then return end

	local source, amount, victim = info.source, info.amount, info.victim

	if not source or source == ParserLib_NONE or not amount or amount == 0 or not victim then
		return
	end

	local unitid
	if source == ParserLib_SELF then
		unitid = "player"
	else
		unitid = roster:GetUnitIDFromName( source )
	end
	if not unitid then -- we're only interested in friendlies
		return 
	end

	local victimid
	if victim == ParserLib_SELF then
		victimid = "player"
	else
		victimid = roster:GetUnitIDFromName( victim )
	end

	local overheal
	if source == ParserLib_SELF then
		overheal = getOverheal( victimid, amount )
		data[playerName] = data[playerName] + amount - overheal
		broadcastdirty = broadcastdirty + amount - overheal
		dirty = true
	else
		if not Violation:UserViolation( source, moduleName ) then
			overheal = getOverheal( victimid, amount )
			self:OnRemoteData( source, amount - overheal )
		end
	end
end

function mod:GetData(player)
	if player then
		return data[player]
	else
		return data
	end
end


