﻿-----------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Parse damage messages from swstats and damagemeters.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- * Problem: Data for users without Violation will be different from Violation
-- user to Violation user because of combat log range.
--   Solution: At all times there is one person in the group with Violation
--   leader status - this status is first determined by who is the group leader
--   or officer, and then randomly. When a leader is selected, he checks what
--   other people in the group have Violation and assign them parsing targets -
--   for example if group 3 has 1 Violation user, that user will report for the
--   rest of the members in group 3. If noone in a group has Violation, a random
--   reporter of first the same class, then the same role, and then completely
--   random will be assigned to parse his data.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Track"] = true,
	["What data types to track."] = true,
	["Tooltip"] = true,
	["What data to show in the tooltip."] = true,
	["Toggle whether to display data from %q in the tooltip."] = true,
	["Colors"] = true,
	["Assign colors to the different data types."] = true,
	["What color to use for the %q data type."] = true,
	["Color the tooltip"] = true,
	["Toggles whether or not to use these colors in the tooltip."] = true,
	
	["Bars"] = true,
	["Settings for the GUI."] = true,
	["Toggle the GUI."] = true,
	["Show"] = true,
	["Display"] = true,
	["What data to show in the GUI."] = true,
	["Texture"] = true,
	["Set the bar texture."] = true,
	["Window Color"] = true,
	["Set the GUI Color"] = true,

	["Resetting data as requested by %s."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,
	["Print on reset"] = true,
	["Prints your current data locally when your data is reset."] = true,

	["Text"] = true,
	["What data to show on the FuBar panel, if available."] = true,

	["Toggle whether to track %q."] = true,
	["Toggle whether to display data from %q on the FuBar panel."] = true,

	["%s has been disabled, since it is not compatible with the latest module API."] = true,
	["No modules running, please enable one or more from the configuration menu."] = true,
	["No data gathered yet."] = true,
	["|cffeda55fClick|r to reset. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to toggle the bar display."] = true,
	["Player"] = true,
	["#"] = true,
	["%s: %d"] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Track"] = "측정",
	["What data types to track."] = "측정할 데이터를 선택합니다.",
	["Tooltip"] = "툴팁",
	["What data to show in the tooltip."] = "툴팁에 표시할 데이터를 선택합니다.",
	["Toggle whether to display data from %q in the tooltip."] = "툴팁에 %q 데이터 표시를 전환합니다.",
	["Colors"] = "색상",
	["Assign colors to the different data types."] = "각각의 데이터 종류에 색상을 지정합니다.",
	["What color to use for the %q data type."] = "%q 데이터에 사용할 색상을 지정합니다.",
	["Color the tooltip"] = "툴팁 색상",
	["Toggles whether or not to use these colors in the tooltip."] = "툴팁 색상 사용을 전환합니다.",

	["Bars"] = "바 설정",
	["Settings for the GUI."] = "GUI 설정",
	["Toggle the GUI."] = "GUI 전환",
	["Show"] = "보기",
	["Display"] = "화면",
	["What data to show in the GUI."] = "화면에 보일 데이터를 선택합니다.",
	["Texture"] = "텍스처",
	["Set the bar texture."] = "바 텍스처를 선택합니다.",

	["Resetting data as requested by %s."] = "%s의 요청에 의해 데이터를 초기화합니다.",
	["Reset"] = "초기화",
	["Options for when to reset the data."] = "데이터 초기화에 대한 설정입니다.",
	["Reset now"] = "즉시 초기화",
	["Immediately resets your data."] = "당신의 데이터를 즉시 초기화합니다.",
	["Reset group"] = "파티 초기화",
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = "모든 파티원에 대한 Violation 데이터를 초기화합니다. 단, 공대장 혹은 승급자 이어야 합니다.",
	["Leader resets"] = "리더 초기화",
	["Accept reset requests from raid leaders or officers."] = "공대장 혹은 오피서로 부터의 초기화 요청을 허용합니다.",
	["Auto-reset"] = "자동 초기화",
	["Entering combat"] = "전투 시작시",
	["Automatically reset the data when entering combat."] = "전투 시작시 자동으로 데이터를 초기화합니다.",
	["Leaving combat"] = "전투 종료시",
	["Automatically reset the data when leaving combat."] = "전투 종료시 자동으로 데이터를 초기화합니다.",
	["Print on reset"] = "초기화 시 출력",
	["Prints your current data locally when your data is reset."] = "당신의 데이터를 초기화할 때 현재 데이터를 출력합니다.",

	["Text"] = "텍스트",
	["What data to show on the FuBar panel, if available."] = "FuBar에 표시할 데이터를 선택합니다.",

	["Toggle whether to track %q."] = "%q 기록 전환",
	["Toggle whether to display data from %q on the FuBar panel."] = "FuBar에 %q 데이터 표시를 전환합니다.",

	["%s has been disabled, since it is not compatible with the latest module API."] = "%s|1이;가; 비활성화 되었습니다, 이후 해당 모듈은 최신 모듈 API와 호환되지 않습니다.",
	["No modules running, please enable one or more from the configuration menu."] = "실행중인 모듈이 없습니다. 환경설정 메뉴에서 하나 이상을 활성화하세요.",
	["No data gathered yet."] = "집계된 데이터가 없습니다.",
	--["|cffeda55fClick|r to reset. |cffeda55fAlt-Click|r to issue a group reset."] = "초기화 하려면 |cffeda55f클릭|r하세요.",
	["Player"] = "플레이어",
	["#"] = "#",
	["%s: %d"] = "%s: %d",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")

local ChatThrottleLib = ChatThrottleLib
if type(ChatThrottleLib) ~= "table" or type(ChatThrottleLib.version) ~= "number" or ChatThrottleLib.version < 13 then
	error("Violation requires ChatThrottleLib 13 or higher.")
end

local SM = AceLibrary("SharedMedia-1.0")

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return key end
		self[key] = "|cff" .. BC:GetHexColor(obj.class) .. key .. "|r"
		return self[key]
	end
})

local RGB = setmetatable({}, {__index = 
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		local r, g, b = BC:GetColor(obj.class)
		self[key] = {r, g, b}
		return self[key]
	end
})

local db = nil

local _G = getfenv(0)
local table_insert = table.insert

local playerName = nil
local percentString = "(%.1f%%)"

local barOffset = 0 -- offset for scrollwheel on the bars.

local moduleVersion = 4

local runningModules = {}
local violationUsers = nil
local mergingInfo = {}
local moduleData = nil
local dataFormatter = {}

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

local printModuleList = {}
local whisperModuleList = {}
local whisperTargets = {
	player = {
		type = "text",
		name = "Player",
		desc = "Whisper the data to the specified player.",
		usage = "<playerName>",
		order = 1,
	},
	spacer = {
		type = "header",
		name = " ",
		order = 50,
	},
}

-----------------------------------------------------------------------
-- Local heap and utility functions
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode="k"})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table_insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

local function Colorize(text, r, g, b)
	return string.format("|cff%02x%02x%02x%s|r", r * 255, g * 255, b * 255, text)
end

-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceDB-2.0",
	"AceConsole-2.0",
	"AceComm-2.0",
	"FuBarPlugin-2.0",
	"AceModuleCore-2.0"
)
local Violation = Violation

Violation:SetModuleMixins("AceEvent-2.0")
Violation.hasIcon = true

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	handler = Violation,
	args = {
		track = {
			type = "group",
			name = L["Track"],
			desc = L["What data types to track."],
			disabled = "~IsActive",
			args = {}, -- Auto-populated by module toggles.
			order = 101,
			pass = true,
			get = function(mod)
				return Violation.db.char.enabledModules[mod]
			end,
			set = function(mod, val)
				Violation.db.char.enabledModules[mod] = val
				Violation:ToggleModuleActive(mod, val)
				Violation:UpdateDisplay()
			end,
		},
		fubar = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {}, -- Auto-populated by module toggles.
			order = 102,
			pass = true,
			get = function(mod)
				return Violation.db.char.textModules[mod]
			end,
			set = function(mod, val)
				Violation.db.char.textModules[mod] = val
				Violation:UpdateText()
			end
		},
		colors = {
			type = "group",
			name = L["Colors"],
			desc = L["Assign colors to the different data types."],
			pass = true,
			get = function(mod)
				if mod == "tooltip" then
					return Violation.db.char.colorTooltip
				else
					local tuple = Violation.db.char.colors[mod]
					return tuple.r, tuple.g, tuple.b
				end
			end,
			set = function(mod, r, g, b)
				if mod == "tooltip" then
					Violation.db.char.colorTooltip = r
					Violation:UpdateTooltip()
				else
					local tuple = Violation.db.char.colors[mod]
					tuple.r = r
					tuple.g = g
					tuple.b = b
					Violation:ScheduleEvent("ColorUpdate", Violation.UpdateDisplay, 5, Violation)
				end
			end,
			args = {
				tooltip = {
					type = "toggle",
					name = L["Color the tooltip"],
					desc = L["Toggles whether or not to use these colors in the tooltip."],
					order = 1,
					disabled = true,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 50,
				},
			}, -- Auto-populated by module toggles.
			order = 103,
		},
		tooltip = {
			type = "group",
			name = L["Tooltip"],
			desc = L["What data to show in the tooltip."],
			args = {},
			order = 104,
			pass = true,
			get = function(mod)
				return Violation.db.char.tooltipModules[mod]
			end,
			set = function(mod, val)
				Violation.db.char.tooltipModules[mod] = val
				Violation:UpdateTooltip()
			end
		},
		bars = {
			type = "group",
			name = L["Bars"],
			desc = L["Settings for the GUI."],
			args = {
				show = {
					type = "toggle",
					name = L["Show"],
					desc = L["Toggle the GUI."],
					get = function() return Violation.db.char.showBars end,
					set = "ToggleBarDisplay",
				},
				display = {
					type = "text",
					name = L["Display"],
					desc = L["What data to show in the GUI."],
					get = function() return Violation.db.char.barModule end,
					set = function(v)
						Violation.db.char.barModule = v
						Violation:OnBarsUpdate()
					end,
					validate = {},
				},
				texture = {
					type = "text",
					name = L["Texture"],
					desc = L["Set the bar texture."],
					get = function() return Violation.db.char.barTexture end,
					set = function(v)
						Violation.db.char.barTexture = v
						Violation:OnBarsUpdate()
					end,
					validate = SM:List("statusbar"),
				},
				color = {
					type = "color",
					name = L["Window Color"],
					desc = L["Set the GUI Color"],
					get = function()
						local tuple = Violation.db.char.windowColor
						return tuple.r, tuple.g, tuple.b
					end,
					set = function(r, g, b)
						local tuple = Violation.db.char.windowColor
						tuple.r = r
						tuple.g = g
						tuple.b = b
						Violation.anchor:SetBackdropBorderColor( r, g, b )
					end,
				},
			},
			order = 105,
		},
		merge = {
			type = "toggle",
			name = "Merge pets",
			desc = "Merge pet data with master.",
			get = function() return Violation.db.char.mergePets end,
			set	= function(value) 
				Violation.db.char.mergePets = value
				Violation:UpdateDisplay()
				Violation:OnBarsUpdate()
			end,
			order = 110,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		resetPrint = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = function() return Violation.db.char.printOnReset end,
			set = function(v) Violation.db.char.printOnReset = v end,
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = "Reset",
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = "ResetGroup",
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = function() return Violation.db.profile.leaderReset end,
					set = function(v) Violation.db.profile.leaderReset = v end,
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_DISABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						Violation.db.profile.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation:IsEventRegistered("PLAYER_REGEN_ENABLED") end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						Violation.db.profile.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
		print = {
			type = "group",
			name = "Print",
			desc = "Prints your current data to a channel.",
			args = {
				say = {
					type = "group",
					name = "Say",
					desc = "Say data.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "SAY") end,
					args = printModuleList,
				},
				guild = {
					type = "group",
					name = "Guild",
					desc = "Print module data on guild channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "GUILD") end,
					args = printModuleList,
				},
				party = {
					type = "group",
					name = "Party",
					desc = "Print module data on guild channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "PARTY") end,
					args = printModuleList,
				},
				raid = {
					type = "group",
					name = "Raid",
					desc = "Print module data on raid channel.",
					pass = true,
					func = function(modName) Violation:PrintModuleDataToChannel(modName, "RAID") end,
					args = printModuleList,
				},
				player = {
					type = "group",
					name = "Whisper",
					desc = "Whisper the data to someone.",
					args = whisperModuleList,
					order = 2,
				},
				limit = {
					type = "range",
					name = "Lines",
					desc = "Maximum number of lines to print.",
					max = 40,
					min = 1,
					step = 1,
					get = function() return Violation.db.char.maxPrintLines end,
					set = function(v) Violation.db.char.maxPrintLines = v end,
					order = 3,
				},
			},
			order = 202,
		},
	},
}

-- options for the rightclick bars menu
local baroptions = options.args.bars

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Violation.modulePrototype.version = -1

local incompatible = {}
local registered = {}
function Violation.modulePrototype:OnInitialize()
	if type(self.name) ~= "string" then
		error("A module is missing its name field.")
	elseif type(self.version) ~= "number" then
		error(self.name .. " doesn't have a version field.")
	elseif registered[self.name] then
		error("A module has been initialized twice for some reason.")
	end

	if self.version == moduleVersion then
		Violation:OnRegister(self)
		registered[self.name] = true
	else
		incompatible[self] = true
		Violation:ModuleIncompatible(self)
	end
end

function Violation.modulePrototype:OnEnable(first)
	if incompatible[self] or not Violation.db.char.enabledModules[self.name] then return end

	if not runningModules[self.name] then
		-- Always reset our data when we enable.
		self:OnReset()

		if type(self.OnModuleEnable) == "function" then
			self:OnModuleEnable()
		end

		runningModules[self.name] = self

		-- Violation:SendCommMessage("GROUP", "ModuleEnabled", self.name)
		Violation:TriggerEvent("Violation_ModuleEnabled",self.name)
	end
end

function Violation.modulePrototype:OnDisable()
	if runningModules[self.name] then
		if type(self.OnModuleDisable) == "function" then
			self:OnModuleDisable()
		end

		runningModules[self.name] = nil

		-- Violation:SendCommMessage("GROUP", "ModuleDisabled", self.name)
		Violation:TriggerEvent("Violation_ModuleDisabled",self.name)
	end
end

function Violation.modulePrototype:new(...) return new(...) end
function Violation.modulePrototype:del(t) return del(t) end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IViolation = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	-- OnModuleDisable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu, for
	-- example "Damage done".
	GetDisplayName = "function",

	-- Must return the module name as it should be displayed in a condensed
	-- display, like "PlayerDPS" should return "DPS".
	GetShortDisplayName = "function",

	-- Must return whether the modules broadcastdata is dirty or not. If it is, a
	-- broadcast will be scheduled by the core.
	IsBroadcastDirty = "function",

	-- Must return wether the modules data is dirty or not. If it is a
	-- diplay update will be scheduled by the core.
	IsDirty = "function",
	
	-- This function is must reset the modules dirty flag that is returned by IsDirty.
	ResetDirty = "function",
	
	-- Must return the data that should be broadcasted to the other players.
	-- Must return 1 number only. This function should reset the broadcastdirty flag as well.
	GetBroadcastData = "function",

	-- Invoked when remote data is received for this module. Usually invoked by
	-- the core, but can also be invoked by other modules when they parse data
	-- from other addons, for example.
	-- Will only get two arguments, the first will be the person the data
	-- belongs to and the second will be a number.
	OnRemoteData = "function",

	-- Invoked by the core when all data should be reset.
	OnReset = "function",

	-- Must return a universal unique ID that will be used for this module when
	-- sending data (where the universe is Violation)
	-- (Not used now, we use a memoization of the module name instead.
	-- GetUUID = "function",

	-- Must return the current dataset for this module. If an argument is given,
	-- it can be assumed to be a player name, and only the data (a number) for
	-- that player (or nil) must be returned.
	GetData = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 37182 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB", "ViolationPerCharDB", "char")
	self:RegisterDefaults("char", {
		enabledModules = {
			["*"] = true,
		},
		textModules = {
			DPS = true,
		},
		tooltipModules = {
			DPS = true,
		},
		colors = {
			["*"] = {
				r = 1, g = 1, b = 1,
			}
		},
		colorTooltip = true,
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		printOnReset = true,
		barModule = "Damage",
		nrBars = 10,
		barTotalHeight = 200,
		barTotalWidth = 154,
		barTexture = "Blizzard",
		showBars = true,
		mergePets = true,
		windowColor = {
			r = 1, g = 0, b = 0,
		}
	})

	self.OnMenuRequest = options

	self:RegisterChatCommand("/violation", self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	self:SetIcon(grey)

	playerName = UnitName("player")
end

function Violation:OnEnable(first)
	db = self.db.char

	if first then
		local memoizations = new()
		for name in self:IterateModules() do
			table_insert(memoizations, name)
		end
		table_insert(memoizations, "LeaderReset")
		table_insert(memoizations, "ViolationPing")
		table_insert(memoizations, "ViolationPong")
		self:RegisterMemoizations(memoizations)
		memoizations = del(memoizations)
	end

	-- Enable/disable all modules.
	for name, mod in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

	barOffset = 0
	self:CreateBarAnchor()
	
	self:Reset()

	self:RegisterComm("Violation", "GROUP", "OnComm")
	self:RegisterComm("Violation", "WHISPER", "OnPrivateComm")

	self:ScheduleRepeatingEvent("Broadcast", self.Broadcast, 5, self)
	self:ScheduleRepeatingEvent("DisplayUpdate", self.UpdateDisplayIfDirty, 2, self)

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:RegisterBucketEvent("RosterLib_RosterUpdated", 2, function()
		self:SendCommMessage("GROUP", "ViolationPing")
		self:UpdatePartyMenu()
		self:UpdateMergingInfo()
	end)

	self:SetIcon(red)
end

function Violation:OnDisable()
	-- Disable all modules.
	for name, mod in pairs(runningModules) do
		self:ToggleModuleActive(name, false)
	end

	self:SetIcon(grey)
end

function Violation:Reset()
	for name, mod in pairs(runningModules) do
		if db.printOnReset then
			local data = mod:GetData(playerName)
			if type(data) == "number" and data > 0 then
				local name = mod:GetDisplayName()
				local color = db.colors[name]
				self:Print(Colorize(string.format(L["%s: %d"], name, data), color.r, color.g, color.b))
			end
		end
		mod:OnReset()
	end
	self:UpdateDisplay()
	self:OnBarsUpdate()
end

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------

function Violation:OnRegister(mod)
	if not AceOO.inherits(mod, IViolation) then
		error(string.format("The module %q does not inherit from the IViolation interface.", tostring(mod.name)))
	end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()

	options.args.track.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to track %q."], displayName),
	}

	options.args.fubar.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q on the FuBar panel."], displayName),
	}

	options.args.tooltip.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = string.format(L["Toggle whether to display data from %q in the tooltip."], displayName),
	}

	options.args.colors.args[moduleName] = {
		type = "color",
		name = displayName,
		desc = string.format(L["What color to use for the %q data type."], displayName),
		order = 100,
	}

	options.args.bars.args.display.validate[moduleName] = displayName

	printModuleList[moduleName] = {
		name = displayName,
		desc = "Send "..displayName.." data.",
		type = "execute",
		disabled = function() return not runningModules[moduleName] end,
	}

	whisperModuleList[moduleName] = {
		type = "group",
		name = displayName,
		desc = "Send "..displayName.." data to player.",
		pass = true,
		get = false,
		set = function(passValue, playerName) return self:PrintModuleDataToChannel(moduleName, "WHISPER", playerName) end,
		func = function(playerName) return self:PrintModuleDataToChannel(moduleName, "WHISPER", playerName) end,
		args = whisperTargets,
		disabled = function() return not runningModules[moduleName] end,
	}
	
	if mod.dataFormatter then
		if type(mod.dataFormatter) == "function" then
			dataFormatter[mod.name] = mod.dataFormatter
		elseif type(mod.dataFormatter) == "string" then
			dataFormatter[mod.name] = function(v) return mod.dataFormatter:format(v) end
		end
	end

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

function Violation:GetModuleVersion()
	return moduleVersion
end

function Violation:ModuleIncompatible(mod)
	self:ToggleModuleActive(mod, false)
	self:Print(string.format(L["%s has been disabled, since it is not compatible with the latest module API."], mod.name))
end

-----------------------------------------------------------------------
-- Pet merging stuff
-----------------------------------------------------------------------

function Violation:UpdateMergingInfo()
	-- Update mergingInfo for named pets (warlocks' and hunters' ones)

	local seen = new()
	local dirty = false
	
	-- Scan roster
	for unit in roster:IterateRoster() do
		-- Gotya
		seen[unit.name] = true
		
		-- Record hunter and warlock pets
		if unit.class == "WARLOCK" or unit.class == "HUNTER" then
			
			-- Check for current pet
			local pet = roster:GetUnitObjectFromUnit(roster:GetPetFromOwner(unit.unitid))
			if pet then
				if not mergingInfo[unit.name] then
					mergingInfo[unit.name] = new()
				end
				if not mergingInfo[unit.name][pet.name] then
					-- This is very ugly, but needed...
					mergingInfo[unit.name][pet.name] = true
					table.insert(mergingInfo[unit.name], pet.name)
					dirty = true
				end
			end
			
		-- Ignore pets of other classes
		elseif mergingInfo[name] then
			mergingInfo[name] = del(mergingInfo[name])
			dirty = true
		end
	end
	
	-- Remove old data
	for name in pairs(mergingInfo) do
		if not seen[name] then
			mergingInfo[name] = del(mergingInfo[name])
			dirty = true
		end
	end
	
	seen = del(seen)
	
	if dirty then
		self:UpdateDisplay()
		self:OnBarsUpdate()
	end
	
end

	
function Violation:GetModuleData(modName, charName)
	local mod = runningModules[modName]
	if not mod then
		return
	end	
	if db.mergePets and mod.canMergeData then
		-- Specific player data
		if charName then
			-- Merge if necessary
			if mergingInfo[charName] then
				return mod:GetMergedData(charName, unpack(mergingInfo[charName]))
			end
		-- All module data
		else
			local seen = new()
			-- Clean previous data table
			moduleData = del(moduleData)
			moduleData = new()
			
			-- Clone module data
			for k,v in pairs(mod:GetData()) do
				moduleData[k] = v
			end
				
			-- Merge data where needed
			for main, others in pairs(mergingInfo) do
				-- Delete pets from moduleData
				for k, v in pairs( others ) do
					moduleData[v] = nil
				end
				moduleData[main] = mod:GetMergedData(main, unpack(others))
			end
			
			return moduleData
		end
		
	end
	-- Fallback to old, plain data
	return mod:GetData(charName)
end


function Violation.modulePrototype:GetMergedData(...)
	local sum = 0
	for i = 1,select("#", ...) do
		sum = sum + (self:GetData(select(i, ...)) or 0)
	end
	return sum
end

-----------------------------------------------------------------------
-- 
-----------------------------------------------------------------------

local function Sorter(a, b) return a[2] > b[2] end

local function PublishModuleData(modName, rowCallback, count, offset)
	local mod = runningModules[modName]
	if not mod then
		return
	end
	local data = Violation:GetModuleData(modName)
	assert(type(data) == "table", "Some module provided us with nil data.")
	
	local dataInfo = new()
	local tbl = new()
	dataInfo.total = 0
	dataInfo.maxValue = 0
	for k, v in pairs(data) do
		if type(v) == "number" and v > 0 then
			dataInfo.total = dataInfo.total + v
			dataInfo.maxValue = math.max(dataInfo.maxValue, v)
			table_insert(tbl, new(k, v))
		end
	end
	
	local tblCount = #tbl
	
	count = math.min(count or tblCount, tblCount)
	offset = math.max(math.min(offset or 0, tblCount-count), 0)

	local n = 0
	if tblCount > 0 then
		local percentage = nil
		
		if tblCount > 1 then
			table.sort(tbl, Sorter)

			if not mod.hidePercentage then
				percentage = new()
				for i, v in ipairs(tbl) do
					table_insert(percentage, (v[2] * 100) / dataInfo.total)
				end
			end
		end
		
		dataInfo.hasMergedData = db.mergePets and mod.canMergeData and next(mergingInfo) and true
		dataInfo.rowCount = tblCount
		dataInfo.module = mod
		dataInfo.formatter = dataFormatter[modName] or tostring
		dataInfo.hasPercentage = percentage
		
		for i = 1, count do
			local rank = offset + i
			local name, value = unpack(tbl[rank])
			if not name then
				break
			end
			rowCallback(i, rank, name, value, percentage and percentage[rank], dataInfo)
			n = n + 1
		end
		
		percentage = del(percentage)
	end	
	
	tbl = del(tbl)
	
	-- We don't want to del tables contained in dataInfo
	-- We really don't want to...
	for k,v in pairs(dataInfo) do
		if type(v) == "table" then
			dataInfo[k] = nil
		end
	end
	
	dataInfo = del(dataInfo)
	
	return n, offset
end

-----------------------------------------------------------------------
-- Channel output
-----------------------------------------------------------------------

function Violation:PrintModuleDataToChannel(modName, channel, target)

	local function doSend(msg, ...)
		if select("#", ...) > 0 then
			msg = msg:format(...)
		end
		ChatThrottleLib:SendChatMessage("BULK", nil, msg, channel, nil, target)
	end

	PublishModuleData(
		modName,
		function(i, rank, name, value, percentage, dataInfo)
			if not dataInfo.header then
				local name = dataInfo.module:GetDisplayName()
				if dataInfo.hasMergedData then
					doSend("==== %s (pets merged) ====", name)
				else
					doSend("==== %s ====", name)
				end
				dataInfo.header = true
			end
			local value = dataInfo.formatter(value)
			if percentage then
				doSend("%d. %s - %s "..percentString, rank, name, value, percentage)
			else
				doSend("%d. %s - %s", rank, name, value)
			end
		end,
		db.maxPrintLines
	)

end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Violation:UpdatePartyMenu()

	local seen = new()
	
	for unit in roster:IterateRoster() do
		local name = unit.name
		seen[name] = true
		if not whisperTargets[name] then
			local option = new()
			option.type = "execute"
			option.name = name
			option.desc = string.format("Send module data to %s.", name)
			option.passValue = name
			option.order = 100
			whisperTargets[name] = option
		end
	end
	
	for k,o in pairs(whisperTargets) do
		if o.passValue and not seen[o.passValue] then
			whisperTargets[k] = del(whisperTargets[k])
		end
	end
	
end

function Violation:OnClick()
	if IsAltKeyDown() then
		self:ResetGroup()
	elseif IsShiftKeyDown() then
		if Violation.anchor:IsShown() then
			self:ToggleBarDisplay(false)
		else
			self:ToggleBarDisplay(true)
		end
	else
		self:Reset()
	end
end

function Violation:UpdateDisplayIfDirty()
	local dirty = nil
	local dirtpile = new()
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] and runningModules[k]:IsDirty() then
			table.insert( dirtpile, k )
			dirty = true
		end
	end
	if dirty then
		self:UpdateDisplay()
	end
	local mod = self:HasModule(db.barModule) and self:GetModule(db.barModule) or nil
	if mod and mod:IsDirty() then
		mod:ResetDirty()
		self:OnBarsUpdate()
	end
	
	for k, v in ipairs( dirtpile ) do
		runningModules[v]:ResetDirty()
	end
	dirtpile = del(dirtpile)
end

local tooltipColors = {
	"child_textR", 1,
	"child_textG", 1,
	"child_textB", 0,
	"child_text2R", 1,
	"child_text2G", 1,
	"child_text2B", 1,
	"child_text3R", 1,
	"child_text3G", 1,
	"child_text3B", 1,
	"child_text4R", 0.65,
	"child_text4G", 0.65,
	"child_text4B", 0.65
}

function Violation:OnTooltipUpdate()
	local no = true
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] then
			no = nil

			PublishModuleData(
				-- module name
				k,
				-- row adding callback
				function(i, rank, name, value, percentage, dataInfo)
					-- Create header on the fly
					if not dataInfo.cat then
						local displayName = dataInfo.module:GetDisplayName()
						local tblCount = dataInfo.rowCount
						dataInfo.cat = tablet:AddCategory(
							"columns", 2 + (tblCount > 1 and 1 or 0) + (dataInfo.hasPercentage and 1 or 0),
							"text", tblCount > 1 and L["#"] or displayName,
							"text2", tblCount > 1 and L["Player"] or "",
							"text3", displayName,
							"child_justify", "LEFT",
							"child_justify2", tblCount > 1 and "LEFT" or "RIGHT",
							"child_justify3", "RIGHT",
							"child_justify4", "RIGHT",
							unpack(tooltipColors)
						)
					end
	
					-- Add row				
					if dataInfo.rowCount > 1 then
						dataInfo.cat:AddLine(
							"text", tostring(rank) .. ".",
							"text2", coloredNames[name],
							"text3", dataInfo.formatter(value or 0),
							"text4", percentage and percentString:format(percentage)
						)
					else
						dataInfo.cat:AddLine(
							"text", coloredNames[name],
							"text2", dataInfo.formatter(value or 0)
						)
					end
				end,
				-- maxcount, offset,
				10, 0
			)
		end
	end
	if no then
		tablet:AddCategory("columns", 1):AddLine("text", L["No modules running, please enable one or more from the configuration menu."])
	end
	tablet:SetHint(L["|cffeda55fClick|r to reset. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to toggle the bar display."])
end

function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v and runningModules[k] then
			local data = self:GetModuleData(k, playerName)
			if type(data) == "number" and data > 0 then
				local shortName = runningModules[k]:GetShortDisplayName()
				local color = db.colors[k]
				data = (dataFormatter[k] or tostring)(data)
				local text = Colorize(string.format("%s: %s", shortName, data), color.r, color.g, color.b)
				if not hasText then hasText = new(text)
				else table_insert(hasText, text) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(table.concat(hasText, " "))
		hasText = del(hasText)
	end
end

-----------------------------------------------------------------------
-- Bar GUI - can eventually be separated by someone else
-----------------------------------------------------------------------

local cachedBars = {}
local currentBars = {}

local function NewBar()
	local f
	if #cachedBars > 0 then
		f = table.remove(cachedBars)
	else
		f = CreateFrame("Button", nil, Violation.anchor)
	end

	-- f:SetWidth( 150 )
	f:SetHeight( 15 )

	if not f.statusbar then
		f.statusbar = CreateFrame("StatusBar", nil, f)
	end
	f.statusbar:ClearAllPoints()
	f.statusbar:SetAllPoints( f ) 
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", db.barTexture) )
	f.statusbar:SetStatusBarColor(.5, .5, .5, 1)
	f.statusbar:SetMinMaxValues(0,100)
	f.statusbar:SetValue(0)

	if not f.lefttext then
		f.lefttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.lefttext:SetFontObject(GameFontHighlightSmall)
	f.lefttext:SetPoint("LEFT", f.statusbar, "LEFT", 0, 0)
	f.lefttext:SetJustifyH("LEFT")
	f.lefttext:SetText("")
	f.lefttext:SetTextColor(1,1,1,1)

	if not f.righttext then
		f.righttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.righttext:SetFontObject(GameFontHighlightSmall)
	f.righttext:SetPoint("RIGHT", f.statusbar, "RIGHT", 0, 0)
	f.righttext:SetJustifyH("RIGHT")
	f.righttext:SetText("")
	f.righttext:SetTextColor(1, 1, 1, 1)

	return f
end

local function SetBar( count, rank, name, textValue, value, vmax, percent)
	if not currentBars[count] then
		currentBars[count] = NewBar()
	end
	local f = currentBars[count]

	f.lefttext:SetText( rank.." "..name )

	if percent then
		f.righttext:SetText( textValue.." "..percentString:format(percent) )
	else
		f.righttext:SetText( textValue )
	end
	if vmax and vmax ~= 0 then
		f.statusbar:SetValue((value/vmax)*100)
	else
		f.statusbar:SetValue(100)
	end
	local color = RGB[name]
	if not color then
		f.statusbar:SetStatusBarColor( .5, .5, .5, 1)
	else
		f.statusbar:SetStatusBarColor( color[1], color[2], color[3], 1)
	end
	if count == 1 then
		f:SetPoint( "TOP", Violation.anchor, "TOP", 0, -32 )
	else
		f:SetPoint( "TOP", currentBars[count-1], "BOTTOM", 0, -1 )
	end
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", db.barTexture) )
	f:SetPoint( "LEFT", Violation.anchor, "LEFT", 2, 0)
	f:SetPoint( "RIGHT", Violation.anchor, "RIGHT", -2, 0)
	f:Show()
end

local function ClearBar( count )
	if not currentBars[count] then return end
	currentBars[count]:Hide()
	table.insert( cachedBars, currentBars[count] )
	currentBars[count] = nil
end

local	function BarAddCallback(i, rank, name, value, percentage, dataInfo)
	return SetBar(i, rank..".", name, dataInfo.formatter(value), value, dataInfo.maxValue, percentage)
end

local function UpdateBars( modName )	
	local barCount, barOffset = PublishModuleData(modName, BarAddCallback, db.nrBars, barOffset)	
	if barCount < db.nrBars then 
		for i = barCount+1, db.nrBars do
			ClearBar( i )
		end
	end
end

function Violation:OnBarsUpdate()
	if not db.showBars then return end
	local mod = db.barModule
	local module = self:GetModule(mod)
	if not module then return end
	self.anchor.header:SetText( module:GetDisplayName() )
	local data = self:GetModuleData(module.name)
	if type(data) == "table" then 
		UpdateBars(module.name)
	end
end

function Violation:OnBarsOffset( up )
	-- we do not check if ouyr barOffset is still valid here. OnBarsUpdate will do that for us
	if up then
		barOffset = barOffset - 1
	else
		barOffset = barOffset + 1
	end
	self:OnBarsUpdate()
end

function Violation:OnBarsSized()
	local f = self.anchor
	
	if not f.isResizing then return end -- don't do anything on frame creation
	
	local y = f:GetHeight()
	local x = f:GetWidth()
	
	db.barTotalHeight = y
	db.barTotalWidth = x
	
	y = math.ceil(y)
	y = y - 34 -- deduct the header size
	
	local nrbars = math.floor( y / 16 )
	if nrbars < 1 then nrbars = 1 end
	
	if nrbars ~= db.nrBars then
		if nrbars < db.nrBars then
			for i = nrbars+1, db.nrBars do
				ClearBar( i )
			end	
		end
		db.nrBars = nrbars
		self:OnBarsUpdate()
	end
end

function Violation:CreateBarAnchor()
	if self.anchor then return end

	local font = GameFontNormal:GetFont()
	
	local f = CreateFrame("Frame", nil, UIParent)
	self.anchor = f
	f.owner = self
	f:Hide()
	f:SetWidth( db.barTotalWidth )
	f:SetHeight( db.barTotalHeight )
	f:SetPoint( "CENTER", UIParent, "CENTER")
	f:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface\\AddOns\\Violation\\textures\\otravi-semi-full-border", edgeSize = 32,
		insets = {left = 1, right = 1, top = 20, bottom = 1},
	})
	f:SetBackdropColor(24/255, 24/255, 24/255)
	f:SetBackdropBorderColor( db.windowColor.r, db.windowColor.g, db.windowColor.b )
	f:EnableMouse(true)
	f:EnableMouseWheel(true)
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(50,50)
	f:RegisterForDrag("LeftButton")
	local dew = AceLibrary("Dewdrop-2.0")
	f:SetScript("OnMouseUp", function(this, a1)
		if a1 == "RightButton" then
			dew:Open( f, "children", function() dew:FeedAceOptionsTable(baroptions) end, "cursorX", true, "cursorY", true )
		end
	end)
	f:SetScript("OnDragStart", function(this) this:StartMoving() end )
	f:SetScript("OnDragStop", function(this) this:StopMovingOrSizing(); this.owner:SavePosition() end )
	f:SetScript("OnSizeChanged", function(this) this.owner:OnBarsSized() end )
	f:SetScript("OnEnter", function(this) this.draghandle:Show() end )
	f:SetScript("OnLeave", function(this) this.owner:ScheduleEvent("ViolationHideDrag", function() Violation.anchor.draghandle:Hide() end, 1) end )
	f:SetScript("OnMouseWheel", function(this, a1) this.owner:OnBarsOffset( a1 > 0 ) end )
	-- header
	f.header = f:CreateFontString(nil,"OVERLAY")
	f.header:SetFont(font, 12)
	f.header:SetText("")
	local module = self:HasModule(db.barModule) and self:GetModule(db.barModule) or nil
	if module then f.header:SetText( module:GetDisplayName() ) end
	f.header:SetJustifyH("LEFT")
	f.header:SetShadowOffset(.8, -.8)
	f.header:SetShadowColor(0, 0, 0, 1)
	f.header:ClearAllPoints()
	f.header:SetPoint("TOPLEFT", f, "TOPLEFT", 7, -14)
	f.header:SetPoint("TOPRIGHT", f, "TOPRIGHT", -28, -14)
	
	-- close button
	f.close = f:CreateTexture(nil, "ARTWORK")
	f.close:SetTexture("Interface\\AddOns\\Violation\\textures\\otravi-close")
	f.close:SetTexCoord(0, .625, 0, .9333)
	f.close:SetWidth(20)
	f.close:SetHeight(14)
	f.close:SetPoint("TOPRIGHT", f, "TOPRIGHT", -7, -15)
	f.closebutton = CreateFrame("Button", nil, f)
	f.closebutton:SetWidth(20)
	f.closebutton:SetHeight(14)
	f.closebutton:SetPoint("CENTER", f.close, "CENTER")
	f.closebutton:SetScript( "OnClick", function() f.owner.db.char.showBars = false; f:Hide() end )
	
	-- drag handle
	
	f.draghandle = CreateFrame("Frame", nil, f)
	f.draghandle:Hide()
	f.draghandle:SetFrameLevel( f:GetFrameLevel() + 10 ) -- place this above everything
	f.draghandle:SetWidth(16)
	f.draghandle:SetHeight(16)
	f.draghandle:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -1, 1)
	f.draghandle:EnableMouse(true)
	f.draghandle:SetScript("OnMouseDown", function(this) this:GetParent().isResizing = true; this:GetParent():StartSizing("BOTTOMRIGHT") end )
	f.draghandle:SetScript("OnMouseUp", function(this) this:GetParent():StopMovingOrSizing(); this:GetParent().owner:OnBarsSized(); this:GetParent().isResizing = nil; end )
	f.draghandle:SetScript("OnEnter", function(this) this:GetParent().owner:CancelScheduledEvent("ViolationHideDrag") end )
	f.draghandle:SetScript("OnLeave", function(this) this:Hide() end )
	
	f.draghandle.texture = f.draghandle:CreateTexture(nil,"BACKGROUND")
	f.draghandle.texture:SetTexture("Interface\\AddOns\\Violation\\textures\\draghandle")
	f.draghandle.texture:SetWidth(16)
	f.draghandle.texture:SetHeight(16)
	f.draghandle.texture:SetBlendMode("ADD")
	f.draghandle.texture:SetPoint( "CENTER", f.draghandle, "CENTER", 0, 0)

	self:RestorePosition()
	
	if db.showBars then
		f:Show()
	end
end

function Violation:SavePosition()
	local f = self.anchor
	local s = f:GetEffectiveScale()

	db.barPosX = f:GetLeft() * s
	db.barPosY = f:GetTop() * s
end

function Violation:RestorePosition()
	local x = db.barPosX
	local y = db.barPosY

	if not x or not y then return end

	local f = self.anchor
	local s = f:GetEffectiveScale()

	f:ClearAllPoints()
	f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x / s, y / s)
end


-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------
function Violation:Broadcast()
	local tbl = nil
	for name, mod in pairs(runningModules) do
		if mod:IsBroadcastDirty() then
			if not tbl then tbl = new() end
			table_insert(tbl, name)
			table_insert(tbl, mod:GetBroadcastData())
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "B", unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		self:Print("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
	self:Reset()
end

function Violation:OnPrivateComm(prefix, sender, distribution, command, ...)
	if distribution ~= "WHISPER" or type(command) ~= "string" then return end
	if command == "ViolationPong" then
		if type(violationUsers) ~= "table" then
			violationUsers = new()
		end
		if violationUsers[sender] then
			violationUsers[sender] = del(violationUsers[sender])
		end
		violationUsers[sender] = new()
		violationUsers[sender]["version"] = select(1,...)
		violationUsers[sender]["revision"] = select(2,...)
		local total = select("#", ...)
		if total > 2 then
			local i, name
			for i = 3, total do
				name = select(i, ... )
				violationUsers[sender][name] = true
			end
		end
	end
end

function Violation:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "B" then
		if type(violationUsers) ~= "table" then
			violationUsers = new()
		end
		if type(violationUsers[sender]) ~= "table" then
			violationUsers[sender] = new()
		end
		local total = select("#", ...)
		local i
		for i = 1, total do
			local name = select(i, ...)
			violationUsers[sender][name] = true
			if runningModules[name] then
				runningModules[name]:OnRemoteData(sender, select(i + 1, ...))
			end
			i = i + 1
			if i > total then break end
		end
	elseif command == "LeaderReset" and db.leaderReset then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj.rank > 0 then
			self:Print(string.format(L["Resetting data as requested by %s."], sender))
			self:Reset(sender)
		end
	elseif command == "ViolationPing" then
		local tbl = new()
		for k in pairs(runningModules) do
			table_insert(tbl, k)
		end
		self:SendCommMessage("WHISPER", sender, "ViolationPong", self.version, self.revision, unpack(tbl))
		tbl = del(tbl)
	end
end

function Violation:UserViolation(user, mod)
	if type(violationUsers) ~= "table" or type(violationUsers[user]) ~= "table" then return end
	if mod and violationUsers[user][mod] then
		return true
	end
	return false
end

function Violation:ToggleBarDisplay(v)
	Violation.db.char.showBars = v
	if v then
		Violation.anchor:Show()
		Violation:OnBarsUpdate()
	else
		Violation.anchor:Hide()
	end
end
